const char net_box_lua[] =
"-- net_box.lua (internal file)\n"
"\n"
"local internal = require 'net.box.lib'\n"
"local msgpack = require 'msgpack'\n"
"local fiber = require 'fiber'\n"
"local socket = require 'socket'\n"
"local log = require 'log'\n"
"local errno = require 'errno'\n"
"local ffi = require 'ffi'\n"
"local yaml = require 'yaml'\n"
"local urilib = require 'uri'\n"
"local buffer = require 'buffer'\n"
"\n"
"-- packet codes\n"
"local OK                = 0\n"
"local SELECT            = 1\n"
"local INSERT            = 2\n"
"local REPLACE           = 3\n"
"local UPDATE            = 4\n"
"local DELETE            = 5\n"
"local CALL              = 6\n"
"local AUTH              = 7\n"
"local EVAL              = 8\n"
"local UPSERT            = 9\n"
"local PING              = 64\n"
"local ERROR_TYPE        = 65536\n"
"\n"
"-- packet keys\n"
"local TYPE              = 0x00\n"
"local SYNC              = 0x01\n"
"local SCHEMA_ID         = 0x05\n"
"local SPACE_ID          = 0x10\n"
"local INDEX_ID          = 0x11\n"
"local LIMIT             = 0x12\n"
"local OFFSET            = 0x13\n"
"local ITERATOR          = 0x14\n"
"local INDEX_BASE        = 0x15\n"
"local KEY               = 0x20\n"
"local TUPLE             = 0x21\n"
"local FUNCTION_NAME     = 0x22\n"
"local USER              = 0x23\n"
"local EXPR              = 0x27\n"
"local OPS               = 0x28\n"
"local DATA              = 0x30\n"
"local ERROR             = 0x31\n"
"local GREETING_SIZE     = 128\n"
"\n"
"local TIMEOUT_INFINITY  = 500 * 365 * 86400\n"
"\n"
"local sequence_mt = { __serialize = 'sequence' }\n"
"local mapping_mt = { __serialize = 'mapping' }\n"
"\n"
"local CONSOLE_FAKESYNC  = 15121974\n"
"local CONSOLE_DELIMITER = \"$EOF$\"\n"
"\n"
"local VSPACE_ID = 281\n"
"local VINDEX_ID = 289\n"
"\n"
"ffi.cdef [[\n"
"enum {\n"
"    /* Maximal length of protocol name in handshake */\n"
"    GREETING_PROTOCOL_LEN_MAX = 32,\n"
"    /* Maximal length of salt in handshake */\n"
"    GREETING_SALT_LEN_MAX = 44,\n"
"};\n"
"\n"
"struct greeting {\n"
"    uint32_t version_id;\n"
"    uint32_t salt_len;\n"
"    char protocol[GREETING_PROTOCOL_LEN_MAX + 1];\n"
"    struct tt_uuid uuid;\n"
"    char salt[GREETING_SALT_LEN_MAX];\n"
"};\n"
"\n"
"int\n"
"greeting_decode(const char *greetingbuf, struct greeting *greeting);\n"
"int strcmp(const char *s1, const char *s2);\n"
"]]\n"
"local builtin = ffi.C\n"
"\n"
"local ch_buf = {}\n"
"local ch_buf_size = 0\n"
"\n"
"local function get_channel()\n"
"    if ch_buf_size == 0 then\n"
"        return fiber.channel()\n"
"    end\n"
"\n"
"    local ch = ch_buf[ch_buf_size]\n"
"    ch_buf[ch_buf_size] = nil\n"
"    ch_buf_size = ch_buf_size - 1\n"
"    return ch\n"
"end\n"
"\n"
"local function free_channel(ch)\n"
"    -- return channel to buffer\n"
"    ch_buf[ch_buf_size + 1] = ch\n"
"    ch_buf_size = ch_buf_size + 1\n"
"end\n"
"\n"
"local function one_tuple(tbl)\n"
"    if tbl == nil then\n"
"        return\n"
"    end\n"
"    if #tbl > 0 then\n"
"        return tbl[1]\n"
"    end\n"
"    return\n"
"end\n"
"\n"
"local requests = {\n"
"    [PING]    = internal.encode_ping;\n"
"    [AUTH]    = internal.encode_auth;\n"
"    [CALL]    = internal.encode_call;\n"
"    [EVAL]    = internal.encode_eval;\n"
"    [INSERT]  = internal.encode_insert;\n"
"    [REPLACE] = internal.encode_replace;\n"
"    [DELETE] = internal.encode_delete;\n"
"    [UPDATE]  = internal.encode_update;\n"
"    [UPSERT]  = internal.encode_upsert;\n"
"    [SELECT]  = function(wbuf, sync, schema_id, spaceno, indexno, key, opts)\n"
"        if opts == nil then\n"
"            opts = {}\n"
"        end\n"
"        if spaceno == nil or type(spaceno) ~= 'number' then\n"
"            box.error(box.error.NO_SUCH_SPACE, '#'..tostring(spaceno))\n"
"        end\n"
"\n"
"        if indexno == nil or type(indexno) ~= 'number' then\n"
"            box.error(box.error.NO_SUCH_INDEX, indexno, '#'..tostring(spaceno))\n"
"        end\n"
"\n"
"        local limit, offset\n"
"        if opts.limit ~= nil then\n"
"            limit = tonumber(opts.limit)\n"
"        else\n"
"            limit = 0xFFFFFFFF\n"
"        end\n"
"        if opts.offset ~= nil then\n"
"            offset = tonumber(opts.offset)\n"
"        else\n"
"            offset = 0\n"
"        end\n"
"        local iterator = require('box.internal').check_iterator_type(opts,\n"
"            key == nil or (type(key) == 'table' and #key == 0))\n"
"\n"
"        internal.encode_select(wbuf, sync, schema_id, spaceno, indexno,\n"
"            iterator, offset, limit, key)\n"
"    end;\n"
"}\n"
"\n"
"local function check_if_space(space)\n"
"    if type(space) == 'table' and space.id ~= nil then\n"
"        return\n"
"    end\n"
"    error(\"Use space:method(...) instead space.method(...)\")\n"
"end\n"
"\n"
"local function space_metatable(self)\n"
"    return {\n"
"        __index = {\n"
"            insert  = function(space, tuple)\n"
"                check_if_space(space)\n"
"                return self:_insert(space.id, tuple)\n"
"            end,\n"
"\n"
"            replace = function(space, tuple)\n"
"                check_if_space(space)\n"
"                return self:_replace(space.id, tuple)\n"
"            end,\n"
"\n"
"            select = function(space, key, opts)\n"
"                check_if_space(space)\n"
"                return self:_select(space.id, 0, key, opts)\n"
"            end,\n"
"\n"
"            delete = function(space, key)\n"
"                check_if_space(space)\n"
"                return self:_delete(space.id, key, 0)\n"
"            end,\n"
"\n"
"            update = function(space, key, oplist)\n"
"                check_if_space(space)\n"
"                return self:_update(space.id, key, oplist, 0)\n"
"            end,\n"
"\n"
"            upsert = function(space, tuple_key, oplist)\n"
"                check_if_space(space)\n"
"                return self:_upsert(space.id, tuple_key, oplist, 0)\n"
"            end,\n"
"\n"
"            get = function(space, key)\n"
"                check_if_space(space)\n"
"                local res = self:_select(space.id, 0, key,\n"
"                                    { limit = 2, iterator = 'EQ' })\n"
"                if #res == 0 then\n"
"                    return\n"
"                end\n"
"                if #res == 1 then\n"
"                    return res[1]\n"
"                end\n"
"                box.error(box.error.MORE_THAN_ONE_TUPLE)\n"
"            end\n"
"        }\n"
"    }\n"
"end\n"
"\n"
"local function check_if_index(idx)\n"
"    if type(idx) == 'table' and idx.id ~= nil and type(idx.space) == 'table' then\n"
"        return\n"
"    end\n"
"    error('Use index:method(...) instead index.method(...)')\n"
"end\n"
"\n"
"local function index_metatable(self)\n"
"    return {\n"
"        __index = {\n"
"            select = function(idx, key, opts)\n"
"                check_if_index(idx)\n"
"                return self:_select(idx.space.id, idx.id, key, opts)\n"
"            end,\n"
"\n"
"            get = function(idx, key)\n"
"                check_if_index(idx)\n"
"                local res = self:_select(idx.space.id, idx.id, key,\n"
"                                    { limit = 2, iterator = 'EQ' })\n"
"                if #res == 0 then\n"
"                    return\n"
"                end\n"
"                if #res == 1 then\n"
"                    return res[1]\n"
"                end\n"
"                box.error(box.error.MORE_THAN_ONE_TUPLE)\n"
"            end,\n"
"\n"
"            min = function(idx, key)\n"
"                check_if_index(idx)\n"
"                local res = self:_select(idx.space.id, idx.id, key,\n"
"                    { limit = 1, iterator = 'GE' })\n"
"                if #res > 0 then\n"
"                    return res[1]\n"
"                end\n"
"            end,\n"
"\n"
"            max = function(idx, key)\n"
"                check_if_index(idx)\n"
"                local res = self:_select(idx.space.id, idx.id, key,\n"
"                    { limit = 1, iterator = 'LE' })\n"
"                if #res > 0 then\n"
"                    return res[1]\n"
"                end\n"
"            end,\n"
"\n"
"            count = function(idx, key)\n"
"                check_if_index(idx)\n"
"                local proc = string.format('box.space.%s.index.%s:count',\n"
"                    idx.space.name, idx.name)\n"
"                local res = self:call(proc, key)\n"
"                if #res > 0 then\n"
"                    return res[1][1]\n"
"                end\n"
"            end,\n"
"\n"
"            delete = function(idx, key)\n"
"                check_if_index(idx)\n"
"                return self:_delete(idx.space.id, key, idx.id)\n"
"            end,\n"
"\n"
"            update = function(idx, key, oplist)\n"
"                check_if_index(idx)\n"
"                return self:_update(idx.space.id, key, oplist, idx.id)\n"
"            end,\n"
"\n"
"            upsert = function(idx, tuple_key, oplist)\n"
"                check_if_index(idx)\n"
"                return self:_upsert(idx.space.id, tuple_key, oplist, idx.id)\n"
"            end,\n"
"\n"
"        }\n"
"    }\n"
"end\n"
"\n"
"local errno_is_transient = {\n"
"    [errno.EAGAIN] = true;\n"
"    [errno.EWOULDBLOCK] = true;\n"
"    [errno.EINTR] = true;\n"
"}\n"
"\n"
"local remote = {}\n"
"\n"
"local remote_methods = {\n"
"    new = function(cls, host, port, opts)\n"
"        local self = { _sync = -1 }\n"
"\n"
"        if type(cls) == 'table' then\n"
"            setmetatable(self, getmetatable(cls))\n"
"        else\n"
"            host, port, opts = cls, host, port\n"
"            setmetatable(self, getmetatable(remote))\n"
"        end\n"
"\n"
"\n"
"        -- uri as the first argument\n"
"        if opts == nil then\n"
"            opts = {}\n"
"            if type(port) == 'table' then\n"
"                opts = port\n"
"                port = nil\n"
"            end\n"
"\n"
"            if port == nil then\n"
"\n"
"                local address = urilib.parse(tostring(host))\n"
"                if address == nil or address.service == nil then\n"
"                    box.error(box.error.PROC_LUA,\n"
"                        \"usage: remote:new(uri[, opts] | host, port[, opts])\")\n"
"                end\n"
"\n"
"                host = address.host\n"
"                port = address.service\n"
"\n"
"                opts.user = address.login or opts.user\n"
"                opts.password = address.password or opts.password\n"
"            end\n"
"        end\n"
"\n"
"\n"
"        self.is_instance = true\n"
"        self.host = host\n"
"        self.port = port\n"
"        self.opts = opts\n"
"\n"
"        if self.opts == nil then\n"
"            self.opts = {}\n"
"        end\n"
"\n"
"        if self.opts.user ~= nil and self.opts.password == nil then\n"
"            box.error(box.error.PROC_LUA,\n"
"                \"net.box: password is not defined\")\n"
"        end\n"
"        if self.opts.user == nil and self.opts.password ~= nil then\n"
"            box.error(box.error.PROC_LUA,\n"
"                \"net.box: user is not defined\")\n"
"        end\n"
"\n"
"\n"
"        if self.host == nil then\n"
"            self.host = 'localhost'\n"
"        end\n"
"\n"
"        self.is_run = true\n"
"        self.state = 'init'\n"
"        self.wbuf = buffer.ibuf(buffer.READAHEAD)\n"
"        self.rpos = 1\n"
"        self.rlen = 0\n"
"        self._schema_id = 0\n"
"\n"
"        self.ch = { sync = {}, fid = {} }\n"
"        self.wait = { state = {} }\n"
"        self.timeouts = {}\n"
"\n"
"        fiber.create(function() self:_connect_worker() end)\n"
"        fiber.create(function() self:_read_worker() end)\n"
"        fiber.create(function() self:_write_worker() end)\n"
"\n"
"        if self.opts.wait_connected == nil or self.opts.wait_connected then\n"
"            self:wait_connected()\n"
"        end\n"
"\n"
"        return self\n"
"    end,\n"
"\n"
"    -- sync\n"
"    sync    = function(self)\n"
"        self._sync = self._sync + 1\n"
"        if self._sync >= 0x7FFFFFFF then\n"
"            self._sync = 0\n"
"        end\n"
"        return self._sync\n"
"    end,\n"
"\n"
"    ping    = function(self)\n"
"        if type(self) ~= 'table' then\n"
"            box.error(box.error.PROC_LUA, \"usage: remote:ping()\")\n"
"        end\n"
"        if not self:is_connected() then\n"
"            return false\n"
"        end\n"
"        local res = self:_request(PING, false)\n"
"\n"
"\n"
"        if res == nil then\n"
"            return false\n"
"        end\n"
"\n"
"        if res.hdr[TYPE] == OK then\n"
"            return true\n"
"        end\n"
"        return false\n"
"    end,\n"
"\n"
"    _console = function(self, line)\n"
"        local data = line..CONSOLE_DELIMITER..\"\\n\\n\"\n"
"        ffi.copy(self.wbuf.wpos, data, #data)\n"
"        self.wbuf.wpos = self.wbuf.wpos + #data\n"
"\n"
"        local res = self:_request_raw(EVAL, CONSOLE_FAKESYNC, data, true)\n"
"        return res.body[DATA]\n"
"    end,\n"
"\n"
"    call    = function(self, proc_name, ...)\n"
"        if type(self) ~= 'table' then\n"
"            box.error(box.error.PROC_LUA, \"usage: remote:call(proc_name, ...)\")\n"
"        end\n"
"\n"
"        proc_name = tostring(proc_name)\n"
"\n"
"        local res = self:_request(CALL, true, proc_name, {...})\n"
"        return res.body[DATA]\n"
"\n"
"    end,\n"
"\n"
"    eval    = function(self, expr, ...)\n"
"        if type(self) ~= 'table' then\n"
"            box.error(box.error.PROC_LUA, \"usage: remote:eval(expr, ...)\")\n"
"        end\n"
"\n"
"        expr = tostring(expr)\n"
"        local data = self:_request(EVAL, true, expr, {...}).body[DATA]\n"
"        local data_len = #data\n"
"        if data_len == 1 then\n"
"            return data[1]\n"
"        elseif data_len == 0 then\n"
"            return\n"
"        else\n"
"            return unpack(data)\n"
"        end\n"
"    end,\n"
"\n"
"    is_connected = function(self)\n"
"        return self.state == 'active' or self.state == 'activew'\n"
"    end,\n"
"\n"
"    wait_connected = function(self, timeout)\n"
"        return self:_wait_state(self._request_states, timeout)\n"
"    end,\n"
"\n"
"    timeout = function(self, timeout)\n"
"        if timeout == nil then\n"
"            return self\n"
"        end\n"
"        if self.is_instance then\n"
"            self.timeouts[ fiber.id() ] = timeout\n"
"            return self\n"
"        end\n"
"\n"
"        return {\n"
"            new = function(cls, host, port, opts)\n"
"                if opts == nil then\n"
"                    opts = {}\n"
"                end\n"
"\n"
"                opts.wait_connected = false\n"
"\n"
"                local cn = self:new(host, port, opts)\n"
"\n"
"                if not cn:wait_connected(timeout) then\n"
"                    cn:close()\n"
"                    box.error(box.error.TIMEOUT)\n"
"                end\n"
"                return cn\n"
"            end\n"
"        }\n"
"    end,\n"
"\n"
"\n"
"    close = function(self)\n"
"        if self.state ~= 'closed' then\n"
"            self:_switch_state('closed')\n"
"            self:_error_waiters('Connection was closed')\n"
"            if self.s ~= nil then\n"
"                self.s:close()\n"
"                self.s = nil\n"
"            end\n"
"        end\n"
"    end,\n"
"\n"
"    -- private methods\n"
"    _fatal = function(self, efmt, ...)\n"
"        if self.state == 'error' then\n"
"            return\n"
"        end\n"
"        local emsg = efmt\n"
"        if select('#', ...) > 0 then\n"
"            emsg = string.format(efmt, ...)\n"
"        end\n"
"\n"
"        if self.s ~= nil then\n"
"            self.s:close()\n"
"            self.s = nil\n"
"        end\n"
"\n"
"        log.warn(\"%s:%s: %s\", self.host or \"\", self.port or \"\", tostring(emsg))\n"
"        self.error = emsg\n"
"        self.space = {}\n"
"        self:_switch_state('error')\n"
"        self:_error_waiters(emsg)\n"
"        self.rpos = 1\n"
"        self.rlen = 0\n"
"        self.version_id = nil\n"
"        self.version = nil\n"
"        self.salt = nil\n"
"    end,\n"
"\n"
"    _wakeup_client = function(self, hdr, body)\n"
"        local sync = hdr[SYNC]\n"
"\n"
"        local ch = self.ch.sync[sync]\n"
"        if ch ~= nil then\n"
"            ch.response = { hdr = hdr, body = body }\n"
"            fiber.wakeup(ch.fid)\n"
"        else\n"
"            log.warn(\"Unexpected response %s\", tostring(sync))\n"
"        end\n"
"    end,\n"
"\n"
"    _error_waiters = function(self, emsg)\n"
"        local waiters = self.ch.sync\n"
"        self.ch.sync = {}\n"
"        for sync, channel in pairs(waiters) do\n"
"            channel.response = {\n"
"                hdr = {\n"
"                    [TYPE] = bit.bor(ERROR_TYPE, box.error.NO_CONNECTION),\n"
"                    [SYNC] = sync\n"
"                },\n"
"                body = {\n"
"                    [ERROR] = emsg\n"
"                }\n"
"            }\n"
"            fiber.wakeup(channel.fid)\n"
"        end\n"
"    end,\n"
"\n"
"    _check_console_response = function(self)\n"
"        local docend = \"\\n...\\n\"\n"
"        local resp = ffi.string(self.rbuf.rpos, self.rbuf:size())\n"
"\n"
"        if #resp < #docend or\n"
"                string.sub(resp, #resp + 1 - #docend) ~= docend then\n"
"            return -1\n"
"        end\n"
"\n"
"        local hdr = { [SYNC] = CONSOLE_FAKESYNC, [TYPE] = 0 }\n"
"        local body = { [DATA] = resp }\n"
"\n"
"        self:_wakeup_client(hdr, body)\n"
"        self.rbuf:read(#resp)\n"
"        return 0\n"
"    end,\n"
"\n"
"    _check_binary_response = function(self)\n"
"        while true do\n"
"            if self.rbuf.rpos + 5 > self.rbuf.wpos then\n"
"                return 5 - (self.rbuf.wpos - self.rbuf.rpos)\n"
"            end\n"
"\n"
"            local rpos, len = msgpack.ibuf_decode(self.rbuf.rpos)\n"
"            if rpos + len > self.rbuf.wpos then\n"
"                return len - (self.rbuf.wpos - rpos)\n"
"            end\n"
"\n"
"            local rpos, hdr = msgpack.ibuf_decode(rpos)\n"
"            local body = {}\n"
"\n"
"            if rpos < self.rbuf.wpos then\n"
"                rpos, body= msgpack.ibuf_decode(rpos)\n"
"            end\n"
"            setmetatable(body, mapping_mt)\n"
"\n"
"            self.rbuf.rpos = rpos\n"
"            self:_wakeup_client(hdr, body)\n"
"\n"
"           if self.rbuf:size() == 0 then\n"
"                return 0\n"
"            end\n"
"        end\n"
"    end,\n"
"\n"
"    _switch_state = function(self, state)\n"
"        if self.state == state or state == nil then\n"
"            return\n"
"        end\n"
"        self.state = state\n"
"\n"
"        local list = self.wait.state[ state ] or {}\n"
"        self.wait.state[ state ] = {}\n"
"\n"
"        for _, fid in pairs(list) do\n"
"            if fid ~= fiber.id() then\n"
"                if self.ch.fid[fid] ~= nil then\n"
"                    -- timeout=0 prevents deadlock if a client went away\n"
"                    -- without removing the channel\n"
"                    self.ch.fid[fid]:put(true, 0)\n"
"                    self.ch.fid[fid] = nil\n"
"                end\n"
"            end\n"
"        end\n"
"    end,\n"
"\n"
"    _wait_state = function(self, states, timeout)\n"
"        timeout = timeout or TIMEOUT_INFINITY\n"
"        while timeout > 0 and self:_is_state(states) ~= true do\n"
"            local started = fiber.time()\n"
"            local fid = fiber.id()\n"
"            local ch = get_channel()\n"
"            for state, _ in pairs(states) do\n"
"                self.wait.state[state] = self.wait.state[state] or {}\n"
"                self.wait.state[state][fid] = fid\n"
"            end\n"
"\n"
"            self.ch.fid[fid] = ch\n"
"            ch:get(timeout)\n"
"            self.ch.fid[fid] = nil\n"
"            free_channel(ch)\n"
"\n"
"            for state, _ in pairs(states) do\n"
"                self.wait.state[state][fid] = nil\n"
"            end\n"
"            timeout = timeout - (fiber.time() - started)\n"
"        end\n"
"        return self.state\n"
"    end,\n"
"\n"
"    _connect_worker = function(self)\n"
"        fiber.name('net.box.connector')\n"
"        local connect_states = { init = true, error = true, closed = true }\n"
"        while self:_wait_state(connect_states) ~= 'closed' do\n"
"\n"
"            if self.state == 'error' then\n"
"                if self.opts.reconnect_after == nil then\n"
"                    self:_switch_state('closed')\n"
"                    return\n"
"                end\n"
"                fiber.sleep(self.opts.reconnect_after)\n"
"                if self.state == 'closed' then\n"
"                    return\n"
"                end\n"
"            end\n"
"\n"
"            self:_switch_state('connecting')\n"
"\n"
"            self.s = socket.tcp_connect(self.host, self.port)\n"
"            if self.s == nil then\n"
"                self:_fatal(errno.strerror())\n"
"            else\n"
"\n"
"                -- on_connect\n"
"                self:_switch_state('handshake')\n"
"                local greetingbuf = self.s:read(GREETING_SIZE)\n"
"                if greetingbuf == nil then\n"
"                    self:_fatal(errno.strerror())\n"
"                elseif #greetingbuf ~= GREETING_SIZE then\n"
"                    self:_fatal(\"Can't read handshake\")\n"
"                else\n"
"                    local greeting = ffi.new('struct greeting')\n"
"                    if builtin.greeting_decode(greetingbuf, greeting) ~= 0 then\n"
"                        self:_fatal(\"Can't decode handshake\")\n"
"                    elseif builtin.strcmp(greeting.protocol, 'Lua console') == 0 then\n"
"                        self.version_id = greeting.version_id\n"
"                        -- enable self:console() method\n"
"                        self.console = self._console\n"
"                        self._check_response = self._check_console_response\n"
"                        -- set delimiter\n"
"                        self:_switch_state('schema')\n"
"\n"
"                        local line = \"require('console').delimiter('\"..CONSOLE_DELIMITER..\"')\\n\\n\"\n"
"                        ffi.copy(self.wbuf.wpos, line, #line)\n"
"                        self.wbuf.wpos = self.wbuf.wpos + #line\n"
"\n"
"                        local res = self:_request_raw(EVAL, CONSOLE_FAKESYNC, line, true)\n"
"\n"
"                        if res.hdr[TYPE] ~= OK then\n"
"                            self:_fatal(res.body[ERROR])\n"
"                        end\n"
"                        self:_switch_state('active')\n"
"                    elseif builtin.strcmp(greeting.protocol, 'Binary') == 0 then\n"
"                        self.version_id = greeting.version_id\n"
"                        self.salt = ffi.string(greeting.salt, greeting.salt_len)\n"
"                        self.console = nil\n"
"                        self._check_response = self._check_binary_response\n"
"                        local s, e = pcall(function()\n"
"                            self:_auth()\n"
"                        end)\n"
"                        if not s then\n"
"                            self:_fatal(e)\n"
"                        end\n"
"\n"
"                        self:reload_schema()\n"
"                    else\n"
"                        self:_fatal(\"Unsupported protocol - \"..\n"
"                            ffi.string(greeting.protocol))\n"
"                    end\n"
"                end\n"
"            end\n"
"        end\n"
"    end,\n"
"\n"
"    _auth = function(self)\n"
"        if self.opts.user == nil or self.opts.password == nil then\n"
"            return\n"
"        end\n"
"\n"
"        self:_switch_state('auth')\n"
"\n"
"        local auth_res = self:_request_internal(AUTH,\n"
"            false, self.opts.user, self.opts.password, self.salt)\n"
"\n"
"        if auth_res.hdr[TYPE] ~= OK then\n"
"            self:_fatal(auth_res.body[ERROR])\n"
"        end\n"
"    end,\n"
"\n"
"    -- states wakeup _read_worker\n"
"    _r_states = {\n"
"        active = true, activew = true, schema = true,\n"
"        schemaw = true, auth = true, authw = true,\n"
"        closed = true,\n"
"    },\n"
"    -- states wakeup _write_worker\n"
"    _rw_states = {\n"
"        activew = true, schemaw = true, authw = true,\n"
"        closed = true,\n"
"    },\n"
"    _request_states = {\n"
"        active = true, activew = true, closed = true,\n"
"    },\n"
"\n"
"    _is_state = function(self, states)\n"
"        return states[self.state] ~= nil\n"
"    end,\n"
"\n"
"    reload_schema = function(self)\n"
"        self._schema_id = 0\n"
"        if self.state == 'closed' or self.state == 'error' then\n"
"            self:_fatal('Can not load schema from the state')\n"
"            return\n"
"        end\n"
"\n"
"        self:_switch_state('schema')\n"
"\n"
"        local resp = self:_request_internal(SELECT,\n"
"            true, VSPACE_ID, 0, nil, { iterator = 'ALL' })\n"
"\n"
"        -- set new schema id after first request\n"
"        self._schema_id = resp.hdr[SCHEMA_ID]\n"
"\n"
"        local spaces = resp.body[DATA]\n"
"        resp = self:_request_internal(SELECT,\n"
"            true, VINDEX_ID, 0, nil, { iterator = 'ALL' })\n"
"        local indexes = resp.body[DATA]\n"
"\n"
"        local sl = {}\n"
"\n"
"        for _, space in pairs(spaces) do\n"
"            local name = space[3]\n"
"            local id = space[1]\n"
"            local engine = space[4]\n"
"            local field_count = space[5]\n"
"\n"
"            local s = {\n"
"                id              = id,\n"
"                name            = name,\n"
"                engine          = engine,\n"
"                field_count     = field_count,\n"
"                enabled         = true,\n"
"                index           = {},\n"
"                temporary       = false\n"
"            }\n"
"            if #space > 5 then\n"
"                local opts = space[6]\n"
"                if type(opts) == 'table' then\n"
"                    -- Tarantool >= 1.7.0\n"
"                    s.temporary = not not opts.temporary\n"
"                elseif type(opts) == 'string' then\n"
"                    -- Tarantool < 1.7.0\n"
"                    s.temporary = string.match(opts, 'temporary') ~= nil\n"
"                end\n"
"            end\n"
"\n"
"            setmetatable(s, space_metatable(self))\n"
"\n"
"            sl[id] = s\n"
"            sl[name] = s\n"
"        end\n"
"\n"
"        for _, index in pairs(indexes) do\n"
"            local idx = {\n"
"                space   = index[1],\n"
"                id      = index[2],\n"
"                name    = index[3],\n"
"                type    = string.upper(index[4]),\n"
"                parts   = {},\n"
"            }\n"
"            local OPTS = 5\n"
"            local PARTS = 6\n"
"\n"
"            if type(index[OPTS]) == 'number' then\n"
"                idx.unique = index[OPTS] == 1 and true or false\n"
"\n"
"                for k = 0, index[PARTS] - 1 do\n"
"                    local pktype = index[7 + k * 2 + 1]\n"
"                    local pkfield = index[7 + k * 2]\n"
"\n"
"                    local pk = {\n"
"                        type = string.upper(pktype),\n"
"                        fieldno = pkfield\n"
"                    }\n"
"                    idx.parts[k] = pk\n"
"                end\n"
"            else\n"
"                for k = 1, #index[PARTS] do\n"
"                    local pktype = index[PARTS][k][2]\n"
"                    local pkfield = index[PARTS][k][1]\n"
"\n"
"                    local pk = {\n"
"                        type = string.upper(pktype),\n"
"                        fieldno = pkfield\n"
"                    }\n"
"                    idx.parts[k - 1] = pk\n"
"                end\n"
"                idx.unique = index[OPTS].is_unique and true or false\n"
"            end\n"
"\n"
"            if sl[idx.space] ~= nil then\n"
"                sl[idx.space].index[idx.id] = idx\n"
"                sl[idx.space].index[idx.name] = idx\n"
"                idx.space = sl[idx.space]\n"
"                setmetatable(idx, index_metatable(self))\n"
"            end\n"
"        end\n"
"\n"
"        self.space = sl\n"
"        if self.state ~= 'error' and self.state ~= 'closed' then\n"
"            self:_switch_state('active')\n"
"        end\n"
"    end,\n"
"\n"
"    _read_worker = function(self)\n"
"        fiber.name('net.box.read')\n"
"        self.rbuf = buffer.ibuf(buffer.READAHEAD)\n"
"        while self:_wait_state(self._r_states) ~= 'closed' do\n"
"            if self.s:readable() then\n"
"                local len = self.s:sysread(self.rbuf.wpos, self.rbuf:unused())\n"
"                if len ~= nil then\n"
"                    if len == 0 then\n"
"                        self:_fatal('Remote host closed connection')\n"
"                    else\n"
"                        self.rbuf.wpos = self.rbuf.wpos + len\n"
"\n"
"                        local advance = self:_check_response()\n"
"                        if advance <= 0 then\n"
"                            advance = buffer.READAHEAD\n"
"                        end\n"
"\n"
"                        self.rbuf:reserve(advance)\n"
"                    end\n"
"                elseif errno_is_transient[errno()] ~= true then\n"
"                    self:_fatal(errno.strerror())\n"
"                end\n"
"            end\n"
"        end\n"
"    end,\n"
"\n"
"    _to_wstate = { active = 'activew', schema = 'schemaw', auth = 'authw' },\n"
"    _to_rstate = { activew = 'active', schemaw = 'schema', authw = 'auth' },\n"
"\n"
"    _write_worker = function(self)\n"
"        fiber.name('net.box.write')\n"
"        while self:_wait_state(self._rw_states) ~= 'closed' do\n"
"            while self.wbuf:size() > 0 do\n"
"                local written = self.s:syswrite(self.wbuf.rpos, self.wbuf:size())\n"
"                if written ~= nil then\n"
"                    self.wbuf.rpos = self.wbuf.rpos + written\n"
"                else\n"
"                    if errno_is_transient[errno()] then\n"
"                    -- the write is with a timeout to detect FIN\n"
"                    -- packet on the receiving end, and close the connection.\n"
"                    -- Every second sockets we iterate the while loop\n"
"                    -- and check the connection state\n"
"                        while self.s:writable(1) == 0 and self.state ~= 'closed' do\n"
"                        end\n"
"                    else\n"
"                        self:_fatal(errno.strerror())\n"
"                        break\n"
"                    end\n"
"                end\n"
"            end\n"
"            self.wbuf:reserve(buffer.READAHEAD)\n"
"            self:_switch_state(self._to_rstate[self.state])\n"
"        end\n"
"    end,\n"
"\n"
"    _request = function(self, reqtype, raise, ...)\n"
"        if self.console then\n"
"            box.error(box.error.UNSUPPORTED, \"console\", reqtype)\n"
"        end\n"
"        local fid = fiber.id()\n"
"        if self.timeouts[fid] == nil then\n"
"            self.timeouts[fid] = TIMEOUT_INFINITY\n"
"        end\n"
"\n"
"        local started = fiber.time()\n"
"\n"
"        self:_wait_state(self._request_states, self.timeouts[fid])\n"
"\n"
"        self.timeouts[fid] = self.timeouts[fid] - (fiber.time() - started)\n"
"\n"
"        if self.state == 'closed' then\n"
"            if raise then\n"
"                box.error(box.error.NO_CONNECTION)\n"
"            end\n"
"        end\n"
"\n"
"        if self.timeouts[fid] <= 0 then\n"
"            self.timeouts[fid] = nil\n"
"            if raise then\n"
"                box.error(box.error.TIMEOUT)\n"
"            else\n"
"                return {\n"
"                    hdr = { [TYPE] = bit.bor(ERROR_TYPE, box.error.TIMEOUT) },\n"
"                    body = { [ERROR] = 'Timeout exceeded' }\n"
"                }\n"
"            end\n"
"        end\n"
"\n"
"        return self:_request_internal(reqtype, raise, ...)\n"
"    end,\n"
"\n"
"    _request_raw = function(self, reqtype, sync, request, raise)\n"
"\n"
"        local fid = fiber.id()\n"
"        if self.timeouts[fid] == nil then\n"
"            self.timeouts[fid] = TIMEOUT_INFINITY\n"
"        end\n"
"\n"
"        self:_switch_state(self._to_wstate[self.state])\n"
"\n"
"        local ch = { fid = fid; }\n"
"        self.ch.sync[sync] = ch\n"
"        fiber.sleep(self.timeouts[fid])\n"
"        local response = ch.response\n"
"        self.ch.sync[sync] = nil\n"
"        self.timeouts[fid] = nil\n"
"\n"
"        if response == nil then\n"
"            if raise then\n"
"                box.error(box.error.TIMEOUT)\n"
"            else\n"
"                return {\n"
"                    hdr = { [TYPE] = bit.bor(ERROR_TYPE, box.error.TIMEOUT) },\n"
"                    body = { [ERROR] = 'Timeout exceeded' }\n"
"                }\n"
"            end\n"
"        end\n"
"\n"
"\n"
"        if response.body[DATA] ~= nil and reqtype ~= EVAL then\n"
"            if rawget(box, 'tuple') ~= nil then\n"
"                for i, v in pairs(response.body[DATA]) do\n"
"                    response.body[DATA][i] =\n"
"                        box.tuple.new(response.body[DATA][i])\n"
"                end\n"
"            end\n"
"            -- disable YAML flow output (useful for admin console)\n"
"            setmetatable(response.body[DATA], sequence_mt)\n"
"        end\n"
"\n"
"        return response\n"
"    end,\n"
"\n"
"    _request_internal = function(self, reqtype, raise, ...)\n"
"        while true do\n"
"            local sync = self:sync()\n"
"            local request = requests[reqtype](self.wbuf, sync, self._schema_id, ...)\n"
"            local response = self:_request_raw(reqtype, sync, request, raise)\n"
"            local resptype = response.hdr[TYPE]\n"
"            if resptype == OK then\n"
"                return response\n"
"            end\n"
"            local err_code = bit.band(resptype, bit.lshift(1, 15) - 1)\n"
"            if err_code ~= box.error.WRONG_SCHEMA_VERSION then\n"
"                if raise then\n"
"                    box.error({\n"
"                        code = err_code,\n"
"                        reason = response.body[ERROR]\n"
"                    })\n"
"                end\n"
"                return response\n"
"            end\n"
"\n"
"            --\n"
"            -- Schema has been changed on the remote server. Try to reload\n"
"            -- schema and re-send request again. Please note reload_schema()\n"
"            -- also calls current function which can loop on WRONG_SCHEMA as\n"
"            -- well. This logic may lead to deep recursion if server\n"
"            -- continuously performs DDL for the long time.\n"
"            --\n"
"            self:reload_schema()\n"
"        end\n"
"        -- unreachable\n"
"    end,\n"
"\n"
"    -- private (low level) methods\n"
"    _select = function(self, spaceno, indexno, key, opts)\n"
"        local res = self:_request(SELECT, true, spaceno, indexno, key, opts)\n"
"        return res.body[DATA]\n"
"    end,\n"
"\n"
"    _insert = function(self, spaceno, tuple)\n"
"        local res = self:_request(INSERT, true, spaceno, tuple)\n"
"        return one_tuple(res.body[DATA])\n"
"    end,\n"
"\n"
"    _replace = function(self, spaceno, tuple)\n"
"        local res = self:_request(REPLACE, true, spaceno, tuple)\n"
"        return one_tuple(res.body[DATA])\n"
"    end,\n"
"\n"
"    _delete  = function(self, spaceno, key, index_id)\n"
"        local res = self:_request(DELETE, true, spaceno, index_id, key, index_id)\n"
"        return one_tuple(res.body[DATA])\n"
"    end,\n"
"\n"
"    _update = function(self, spaceno, key, oplist, index_id)\n"
"        local res = self:_request(UPDATE, true, spaceno, index_id, key, oplist)\n"
"        return one_tuple(res.body[DATA])\n"
"    end,\n"
"\n"
"    _upsert = function(self, spaceno, tuple_key, oplist, index_id)\n"
"        local res = self:_request(UPSERT, true, spaceno,\n"
"                                  index_id, tuple_key, oplist)\n"
"        return one_tuple(res.body[DATA])\n"
"    end,\n"
"}\n"
"\n"
"setmetatable(remote, { __index = remote_methods })\n"
"\n"
"local function rollback()\n"
"    if rawget(box, 'rollback') ~= nil then\n"
"        -- roll back local transaction on error\n"
"        box.rollback()\n"
"    end\n"
"end\n"
"\n"
"local function handle_call_result(status, ...)\n"
"    if not status then\n"
"        rollback()\n"
"        return box.error(box.error.PROC_LUA, (...))\n"
"    end\n"
"    if select('#', ...) == 1 and type((...)) == 'table' then\n"
"        local result = (...)\n"
"        for i, v in pairs(result) do\n"
"            result[i] = box.tuple.new(v)\n"
"        end\n"
"        return result\n"
"    else\n"
"        local result = {}\n"
"        for i=1,select('#', ...), 1 do\n"
"            result[i] = box.tuple.new((select(i, ...)))\n"
"        end\n"
"        return result\n"
"    end\n"
"end\n"
"\n"
"local function handle_eval_result(status, ...)\n"
"    if not status then\n"
"        rollback()\n"
"        return box.error(box.error.PROC_LUA, (...))\n"
"    end\n"
"    return ...\n"
"end\n"
"\n"
"remote.self = {\n"
"    ping = function() return true end,\n"
"    reload_schema = function() end,\n"
"    close = function() end,\n"
"    timeout = function(self) return self end,\n"
"    wait_connected = function(self) return true end,\n"
"    is_connected = function(self) return true end,\n"
"    call = function(_box, proc_name, ...)\n"
"        if type(_box) ~= 'table' then\n"
"            box.error(box.error.PROC_LUA, \"usage: remote:call(proc_name, ...)\")\n"
"        end\n"
"        proc_name = tostring(proc_name)\n"
"        local status, proc, obj = pcall(package.loaded['box.internal'].\n"
"            call_loadproc, proc_name)\n"
"        if not status then\n"
"            rollback()\n"
"            return box.error() -- re-throw\n"
"        end\n"
"        local result\n"
"        if obj ~= nil then\n"
"            return handle_call_result(pcall(proc, obj, ...))\n"
"        else\n"
"            return handle_call_result(pcall(proc, ...))\n"
"        end\n"
"    end,\n"
"    eval = function(_box, expr, ...)\n"
"        if type(_box) ~= 'table' then\n"
"            box.error(box.error.PROC_LUA, \"usage: remote:eval(expr, ...)\")\n"
"        end\n"
"        local proc, errmsg = loadstring(expr)\n"
"        if not proc then\n"
"            proc, errmsg = loadstring(\"return \"..expr)\n"
"        end\n"
"        if not proc then\n"
"            rollback()\n"
"            return box.error(box.error.PROC_LUA, errmsg)\n"
"        end\n"
"        return handle_eval_result(pcall(proc, ...))\n"
"    end\n"
"}\n"
"\n"
"setmetatable(remote.self, {\n"
"    __index = function(self, key)\n"
"        if key == 'space' then\n"
"            -- proxy self.space to box.space\n"
"            return require('box').space\n"
"        end\n"
"        return nil\n"
"    end\n"
"})\n"
"\n"
"package.loaded['net.box'] = remote\n"
""
;
