const char snapshot_daemon_lua[] =
"-- snapshot_daemon.lua (internal file)\n"
"\n"
"local log = require 'log'\n"
"local fiber = require 'fiber'\n"
"local fio = require 'fio'\n"
"local yaml = require 'yaml'\n"
"local errno = require 'errno'\n"
"local digest = require 'digest'\n"
"local pickle = require 'pickle'\n"
"\n"
"local PREFIX = 'snapshot_daemon'\n"
"\n"
"local daemon = {\n"
"    snapshot_period = 0;\n"
"    snapshot_count = 6;\n"
"    fiber = nil;\n"
"    control = nil;\n"
"}\n"
"\n"
"local function sprintf(fmt, ...) return string.format(fmt, ...) end\n"
"\n"
"-- create snapshot, return true if no errors\n"
"local function snapshot()\n"
"    log.info(\"making snapshot...\")\n"
"    local s, e = pcall(function() box.snapshot() end)\n"
"    if s then\n"
"        return true\n"
"    end\n"
"    -- don't complain in the log if the snapshot already exists\n"
"    if errno() == errno.EEXIST then\n"
"        return false\n"
"    end\n"
"    log.error(\"error while creating snapshot: %s\", e)\n"
"    return false\n"
"end\n"
"\n"
"-- create snapshot\n"
"local function make_snapshot(last_snap)\n"
"\n"
"    if daemon.snapshot_period == nil then\n"
"        return false\n"
"    end\n"
"\n"
"    if not(daemon.snapshot_period > 0) then\n"
"        return false\n"
"    end\n"
"\n"
"\n"
"    if last_snap == nil then\n"
"        return snapshot()\n"
"    end\n"
"\n"
"    local vclock = box.info.vclock\n"
"    local lsn = 0\n"
"    for i, vlsn in pairs(vclock) do\n"
"        lsn = lsn + vlsn\n"
"    end\n"
"\n"
"    local snap_name = sprintf('%020d.snap', tonumber(lsn))\n"
"    if fio.basename(last_snap) == snap_name then\n"
"        if daemon.last_snap_name ~= snap_name then\n"
"            daemon.last_snap_name = snap_name\n"
"            log.debug('snapshot file %s already exists', last_snap)\n"
"        end\n"
"        return false\n"
"    end\n"
"    daemon.last_snap_name = snap_name\n"
"\n"
"    local snstat = fio.stat(last_snap)\n"
"    if snstat == nil then\n"
"        log.error(\"can't stat %s: %s\", last_snap, errno.strerror())\n"
"        return false\n"
"    end\n"
"    if snstat.mtime <= fiber.time() + daemon.snapshot_period then\n"
"        return snapshot()\n"
"    end\n"
"end\n"
"\n"
"-- check filesystem and current time\n"
"local function process(self)\n"
"    local snaps = fio.glob(fio.pathjoin(box.cfg.snap_dir, '*.snap'))\n"
"\n"
"    if snaps == nil then\n"
"        log.error(\"can't read snap_dir %s: %s\", box.cfg.snap_dir,\n"
"                  errno.strerror())\n"
"        return\n"
"    end\n"
"\n"
"    if not make_snapshot(snaps[#snaps]) then\n"
"        return\n"
"    end\n"
"\n"
"    -- cleanup code\n"
"    if daemon.snapshot_count == nil then\n"
"        return\n"
"    end\n"
"\n"
"    if not (self.snapshot_count > 0) then\n"
"        return\n"
"    end\n"
"\n"
"\n"
"    -- reload snap list after snapshot\n"
"    snaps = fio.glob(fio.pathjoin(box.cfg.snap_dir, '*.snap'))\n"
"    local xlogs = fio.glob(fio.pathjoin(box.cfg.wal_dir, '*.xlog'))\n"
"    if xlogs == nil then\n"
"        log.error(\"can't read wal_dir %s: %s\", box.cfg.wal_dir,\n"
"                  errno.strerror())\n"
"        return\n"
"    end\n"
"\n"
"    while #snaps > self.snapshot_count do\n"
"        local rm = snaps[1]\n"
"        table.remove(snaps, 1)\n"
"\n"
"        log.info(\"removing old snapshot %s\", rm)\n"
"        if not fio.unlink(rm) then\n"
"            log.error(\"error while removing %s: %s\",\n"
"                      rm, errno.strerror())\n"
"            return\n"
"        end\n"
"    end\n"
"\n"
"\n"
"    local snapno = fio.basename(snaps[1], '.snap')\n"
"\n"
"    while #xlogs > 0 do\n"
"        if #xlogs < 2 then\n"
"            break\n"
"        end\n"
"\n"
"        if fio.basename(xlogs[1], '.xlog') > snapno then\n"
"            break\n"
"        end\n"
"\n"
"        if fio.basename(xlogs[2], '.xlog') > snapno then\n"
"            break\n"
"        end\n"
"\n"
"\n"
"        local rm = xlogs[1]\n"
"        table.remove(xlogs, 1)\n"
"        log.info(\"removing old xlog %s\", rm)\n"
"\n"
"        if not fio.unlink(rm) then\n"
"            log.error(\"error while removing %s: %s\",\n"
"                      rm, errno.strerror())\n"
"            return\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function daemon_fiber(self)\n"
"    fiber.name(PREFIX)\n"
"    log.info(\"started\")\n"
"\n"
"    --\n"
"    -- Add random offset to the initial period to avoid simultaneous\n"
"    -- snapshotting when multiple instances of tarantool are running\n"
"    -- on the same host.\n"
"    -- See https://github.com/tarantool/tarantool/issues/732\n"
"    --\n"
"    local random = pickle.unpack('i', digest.urandom(4))\n"
"    local offset = random % self.snapshot_period\n"
"    while true do\n"
"        local period = self.snapshot_period + offset\n"
"        -- maintain next_snapshot_time as a self member for testing purposes\n"
"        self.next_snapshot_time = fiber.time() + period\n"
"        log.info(\"scheduled the next snapshot at %s\",\n"
"                os.date(\"%c\", self.next_snapshot_time))\n"
"        local msg = self.control:get(period)\n"
"        if msg == 'shutdown' then\n"
"            break\n"
"        elseif msg == 'reload' then\n"
"            log.info(\"reloaded\") -- continue\n"
"        elseif msg == nil and box.info.status == 'running' then\n"
"            local s, e = pcall(process, self)\n"
"            if not s then\n"
"                log.error(e)\n"
"            end\n"
"            offset = 0\n"
"        end\n"
"    end\n"
"    self.next_snapshot_time = nil\n"
"    log.info(\"stopped\")\n"
"end\n"
"\n"
"local function reload(self)\n"
"    if self.snapshot_period > 0 then\n"
"        if self.control == nil then\n"
"            -- Start daemon\n"
"            self.control = fiber.channel()\n"
"            self.fiber = fiber.create(daemon_fiber, self)\n"
"            fiber.sleep(0)\n"
"        else\n"
"            -- Reload daemon\n"
"            self.control:put(\"reload\")\n"
"            --\n"
"            -- channel:put() doesn't block the writer if there\n"
"            -- is a ready reader. Give daemon fiber way so that\n"
"            -- it can execute before reload() returns to the caller.\n"
"            --\n"
"            fiber.sleep(0)\n"
"        end\n"
"    elseif self.control ~= nil then\n"
"        -- Shutdown daemon\n"
"        self.control:put(\"shutdown\")\n"
"        self.fiber = nil\n"
"        self.control = nil\n"
"        fiber.sleep(0) -- see comment above\n"
"    end\n"
"end\n"
"\n"
"setmetatable(daemon, {\n"
"    __index = {\n"
"        set_snapshot_period = function()\n"
"            daemon.snapshot_period = box.cfg.snapshot_period\n"
"            reload(daemon)\n"
"            return\n"
"        end,\n"
"\n"
"        set_snapshot_count = function()\n"
"            if math.floor(box.cfg.snapshot_count) ~= box.cfg.snapshot_count then\n"
"                box.error(box.error.CFG, \"snapshot_count\",\n"
"                         \"must be an integer\")\n"
"            end\n"
"            daemon.snapshot_count = box.cfg.snapshot_count\n"
"            reload(daemon)\n"
"        end\n"
"    }\n"
"})\n"
"\n"
"if box.internal == nil then\n"
"    box.internal = { [PREFIX] = daemon }\n"
"else\n"
"    box.internal[PREFIX] = daemon\n"
"end\n"
""
;
