const char init_lua[] =
"-- init.lua -- internal file\n"
"\n"
"local ffi = require('ffi')\n"
"ffi.cdef[[\n"
"struct type;\n"
"struct method;\n"
"struct error;\n"
"\n"
"struct type {\n"
"    const char *name;\n"
"    const struct type *parent;\n"
"    const struct method *methods;\n"
"};\n"
"\n"
"enum {\n"
"    DIAG_ERRMSG_MAX = 512,\n"
"    DIAG_FILENAME_MAX = 256\n"
"};\n"
"\n"
"typedef void (*error_f)(struct error *e);\n"
"\n"
"struct error {\n"
"    error_f _destroy;\n"
"    error_f _raise;\n"
"    error_f _log;\n"
"    const struct type *_type;\n"
"    int _refs;\n"
"    /** Line number. */\n"
"    unsigned _line;\n"
"    /* Source file name. */\n"
"    char _file[DIAG_FILENAME_MAX];\n"
"    /* Error description. */\n"
"    char _errmsg[DIAG_ERRMSG_MAX];\n"
"};\n"
"\n"
"/* TODO: remove these declarations */\n"
"const struct error *\n"
"box_error_last(void);\n"
"const char *\n"
"box_error_message(const struct error *);\n"
"\n"
"double\n"
"tarantool_uptime(void);\n"
"typedef int32_t pid_t;\n"
"pid_t getpid(void);\n"
"]]\n"
"\n"
"local function error_type(err)\n"
"    return ffi.string(err._type.name)\n"
"end\n"
"\n"
"local function error_message(err)\n"
"    return ffi.string(err._errmsg)\n"
"end\n"
"\n"
"local function error_trace(err)\n"
"    if err._file[0] == 0 then\n"
"        return {}\n"
"    end\n"
"    return {\n"
"        { file = ffi.string(err._file), line = tonumber(err._line) };\n"
"    }\n"
"end\n"
"\n"
"-- TODO: Use reflection\n"
"local error_fields = {\n"
"    [\"type\"]        = error_type;\n"
"    [\"message\"]     = error_message;\n"
"    [\"trace\"]       = error_trace;\n"
"}\n"
"\n"
"local function error_unpack(err)\n"
"    if not ffi.istype('struct error', err) then\n"
"        error(\"Usage: error:unpack()\")\n"
"    end\n"
"    local result = {}\n"
"    for key, getter in pairs(error_fields)  do\n"
"        result[key] = getter(err)\n"
"    end\n"
"    return result\n"
"end\n"
"\n"
"local function error_raise(err)\n"
"    if not ffi.istype('struct error', err) then\n"
"        error(\"Usage: error:raise()\")\n"
"    end\n"
"    error(err)\n"
"end\n"
"\n"
"local function error_serialize(err)\n"
"    -- Return an error message only in admin console to keep compatibility\n"
"    return error_message(err)\n"
"end\n"
"\n"
"local error_methods = {\n"
"    [\"unpack\"] = error_unpack;\n"
"    [\"raise\"] = error_raise;\n"
"    [\"__serialize\"] = error_serialize;\n"
"}\n"
"\n"
"local function error_index(err, key)\n"
"    local getter = error_fields[key]\n"
"    if getter ~= nil then\n"
"        return getter(err)\n"
"    end\n"
"    return error_methods[key]\n"
"end\n"
"\n"
"local error_mt = {\n"
"    __index = error_index;\n"
"    __tostring = error_message;\n"
"};\n"
"\n"
"ffi.metatype('struct error', error_mt);\n"
"\n"
"-- Override pcall to support Tarantool exceptions\n"
"local pcall_lua = pcall\n"
"\n"
"local function pcall_wrap(status, ...)\n"
"    if status == true then\n"
"        return true, ...\n"
"    end\n"
"    if ffi.istype('struct error', (...)) then\n"
"        -- Return the Tarantool error as string to keep compatibility.\n"
"        -- Caller should check box.error.last() to get additional information.\n"
"        return false, tostring((...))\n"
"    elseif ... == 'C++ exception' then\n"
"        return false, ffi.string(ffi.C.box_error_message(ffi.C.box_error_last()))\n"
"    end\n"
"    return status, ...\n"
"end\n"
"pcall = function(fun, ...)\n"
"    return pcall_wrap(pcall_lua(fun, ...))\n"
"end\n"
"\n"
"dostring = function(s, ...)\n"
"    local chunk, message = loadstring(s)\n"
"    if chunk == nil then\n"
"        error(message, 2)\n"
"    end\n"
"    return chunk(...)\n"
"end\n"
"\n"
"local function uptime()\n"
"    return tonumber(ffi.C.tarantool_uptime());\n"
"end\n"
"\n"
"local function pid()\n"
"    return tonumber(ffi.C.getpid())\n"
"end\n"
"\n"
"return {\n"
"    uptime = uptime;\n"
"    pid = pid;\n"
"}\n"
""
;
